import logging
from pprint import pformat
from typing import List, Dict, Type, Any

from sync.datasource import DataSource
from sync.mixins import LinkProvider, Timestamped


class InMemoryDs(DataSource):
    def preload_type(self, preload_type, path) -> bool:
        """skip"""
        return True

    def __init__(self):
        self.store: Dict[Type, Dict[int, Any]] = dict()

    def create(self, entity):
        if entity.__class__ not in self.store:
            self.store[entity.__class__] = dict()

        if entity not in self.store[entity.__class__].values():
            next_id = self.next_id(entity.__class__)
            self.store[entity.__class__][next_id] = entity
            return True
        return False

    def all(self, of_type, limit=None) -> set:
        if of_type in self.store:
            return list(self.store[of_type].values()).copy()
        else:
            return list()

    def delete(self, entity) -> bool:
        entity_id = self.id(entity)
        if entity_id in self.store[entity.__class__]:
            del self.store[entity.__class__][entity_id]
            return True
        return False

    def entity_types(self) -> List[type]:
        return self.store.keys()

    @staticmethod
    def from_collection(entities, db=None):
        if not db:
            db = InMemoryDs()

        for element in entities:
            db.create(element)

            for prop in element.__dict__:
                value = element.__dict__[prop]
                if isinstance(value, set) or isinstance(value, list):
                    InMemoryDs.from_collection(value, db)

        return db

    def find(self, by_type, by_conditions=dict()):
        def object_matches(object, conditions):
            if not object:
                return False
            for k, v in conditions:
                if getattr(object, k) != v:
                    return False
            return True

        if by_type not in self.store:
            return []

        result_set = [
            e
            for e in self.store[by_type].values()
            if object_matches(e, by_conditions.items())
        ]

        logging.debug(
            f"Searching for {by_type.__name__} by {by_conditions}: {len(result_set)} found"
        )
        return result_set

    def refresh(self, entity):
        if entity in self.all(entity.__class__):
            return entity
        else:
            return None

    def __repr__(self):
        return pformat(self.store)

    def persist(self, entity):
        self.from_collection([entity], self)

    def get(self, type, id):
        return self.store.get(type).get(int(id))

    def id(self, entity):
        class_store = self.store.get(entity.__class__)
        for _id, check_entity in class_store.items():
            if check_entity == entity:
                return _id
        return None

    def next_id(self, type):
        table = self.store.get(type)
        if not table or len(table.keys()) == 0:
            return 0
        else:
            return list(table.keys())[-1] + 1


class DoubleDictToSet:
    def __init__(self):
        self.dict = dict()

    def add(self, key1, key2, value):
        if key1 not in self.dict:
            self.dict[key1] = dict()
        if key2 not in self.dict[key1]:
            self.dict[key1][key2] = set()
        self.dict[key1][key2].add(value)

    def get(self, key1, key2):
        return self.dict[key1][key2].copy()

    def __contains__(self, item):
        key1, key2 = item
        return key1 in self.dict and key2 in self.dict[key1]


class InMemoryLinkProvider(LinkProvider):
    def __init__(self):
        self.links = DoubleDictToSet()

    def link(self, entity, other, entity_ds: DataSource, other_ds: DataSource):
        other_type = other.__class__
        entity_type = entity.__class__
        if (
                not (entity, other_type) in self.links
                or not (other, entity_type) in self.links
        ):
            self.links.add(entity, other_type, other)
            self.links.add(other, entity_type, entity)
            return True
        return False

    def others(
            self, entity, entity_ds: DataSource, other_type: Type, other_ds: DataSource
    ):
        if (entity, other_type) in self.links:
            return self.links.get(entity, other_type)
        else:
            return None

    def unlink(self, entity, entity_ds: DataSource, other_ds: DataSource):
        pass


class InMemoryTimestamp(Timestamped):
    _updated_at: int = 0

    def updated_at(self):
        return self._updated_at

    def update_timestamp(self, timestamp):
        self._updated_at = timestamp


class Hierarchical:
    @property
    def parent(self):
        if hasattr(self, "_parent"):
            return self._parent
        else:
            return None

    @parent.setter
    def parent(self, new):
        if hasattr(self, "_parent") and self._parent and self in self._parent._children:
            self._parent._children.remove(self)

        if new:
            if not hasattr(new, "_children"):
                new._children = set()
            new._children.add(self)

        self._parent = new

    @property
    def children(self):
        return self._children

    @children.setter
    def children(self, new):
        for child in new:
            child._parent = self
        self._children = new


class InMemoryBase(Hierarchical, InMemoryTimestamp):
    def __repr__(self, show_children=False, show_parent=True, indent=0):
        result = " " * indent
        result += f"{self.__class__.__name__}("

        if hasattr(self, "name"):
            result += f"name={self.name}"

        if show_children and self.children.__len__() > 0:
            result += f", children=[\n"
            for child in self.children:
                result += f"{child.__repr__(indent=(indent + 2))}\n"
            result += " " * indent + "]"
        if show_parent and self.parent:
            result += f", parent={self.parent.__repr__(show_parent=False, show_children=False, indent=indent)}"
        result += ")"
        return result



from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import *
from sqlalchemy.orm import relationship

from demo.util import Repr, UpdatedAt

DesireBase = declarative_base()


class CloudServiceProvider(DesireBase, UpdatedAt, Repr):
    __tablename__ = "cloud_service_providers"
    id = Column(Integer, primary_key=True)
    name = Column(String)
    upgrade_allowed = Column(Boolean)
    children = relationship("CloudRegion", back_populates="parent")
    pending_update = Column(Boolean, default=False, nullable=False)


class CloudRegion(DesireBase, UpdatedAt, Repr):
    __tablename__ = "cloud_regions"
    id = Column(Integer, primary_key=True)
    name = Column(String)
    external_ref = Column(String)
    parent_id = Column(Integer, ForeignKey("cloud_service_providers.id"))
    parent = relationship("CloudServiceProvider", back_populates="children")
    children = relationship("CloudHandover", back_populates="parent")


class CloudHandover(DesireBase, UpdatedAt, Repr):
    __tablename__ = "cloud_handovers"
    id = Column(Integer, primary_key=True)
    name = Column(String)
    external_ref = Column(String)
    parent_id = Column(Integer, ForeignKey("cloud_regions.id"))
    parent = relationship("CloudRegion", back_populates="children")
    children = relationship("NIC", back_populates="parent")


class NIC(DesireBase, UpdatedAt, Repr):
    __tablename__ = "nics"
    id = Column(Integer, primary_key=True)
    name = Column(String)
    external_ref = Column(String)
    parent_id = Column(Integer, ForeignKey("cloud_handovers.id"))
    parent = relationship("CloudHandover", back_populates="children")



from abc import ABC, abstractmethod
from typing import Type

from sync.datasource import DataSource


class Timestamped:
    def updated_at(self):
        pass

    def update_timestamp(self, timestamp):
        pass


class LinkProvider(ABC):
    @abstractmethod
    def link(self, entity, other, entity_ds: DataSource, other_ds: DataSource):
        pass

    @abstractmethod
    def others(
        self, entity, entity_ds: DataSource, other_type: Type, other_ds: DataSource
    ):
        pass

    @abstractmethod
    def unlink(self, entity, entity_ds: DataSource, other_ds: DataSource):
        pass



from typing import Type

from redis.client import Redis

from sync.datasource import DataSource
from sync.mixins import LinkProvider


class RedisLinkProvider(LinkProvider):
    def __init__(self, purge=False, **redis_parameters):
        redis_parameters.update({"decode_responses": True, "encoding": "utf-8"})
        self.redis = Redis(**redis_parameters)
        if purge:
            self.redis.flushdb()

    @staticmethod
    def _generate_link_key(from_type: Type, from_entity_id, to_type: Type):
        return f"{from_type.__name__}:{from_entity_id}:{to_type.__name__}"

    @staticmethod
    def _generate_link_wildcard(from_type: Type, from_entity_id):
        return f"{from_type.__name__}:{from_entity_id}:*"

    @staticmethod
    def _generate_link_value(value_type: Type, value_id):
        return f"{value_type.__name__}:{value_id}"

    def link(self, entity, other, entity_ds: DataSource, other_ds: DataSource):
        self.link_unidirectional(entity, other, entity_ds, other_ds)
        self.link_unidirectional(other, entity, other_ds, entity_ds)
        return True

    def link_unidirectional(
        self, entity, other, entity_ds: DataSource, other_ds: DataSource
    ):
        link_key = RedisLinkProvider._generate_link_key(
            entity.__class__, entity_ds.id(entity), other.__class__
        )
        link_value = RedisLinkProvider._generate_link_value(
            other.__class__, other_ds.id(other)
        )
        self.redis.sadd(link_key, link_value)

    def others(
        self, entity, entity_ds: DataSource, other_type: Type, other_ds: DataSource
    ):
        link_key = RedisLinkProvider._generate_link_key(
            entity.__class__, entity_ds.id(entity), other_type
        )
        values = self.redis.smembers(link_key)
        result = []
        for member_key in values:
            other_type_from_value, value = member_key.split(":")
            if other_type_from_value == other_type.__name__:
                other = other_ds.get(other_type, value)
                result.append(other)

        return result

    def others_by_id(
        self, entity_class, entity_id, other_type: Type, other_ds: DataSource
    ):
        link_key = (
            f"{entity_class.__name__}:{entity_id}:{other_type.__name__}"
        )
        values = self.redis.smembers(link_key)
        result = []
        for member_key in values:
            other_type_from_value, value = member_key.split(":")
            if other_type_from_value == other_type.__name__:
                other = other_ds.get(other_type, value)
                result.append(other)

        return result

    def unlink(self, entity, entity_ds: DataSource, other_ds: DataSource):
        link_wildcard = RedisLinkProvider._generate_link_wildcard(
            entity.__class__, entity_ds.id(entity)
        )
        for key in self.redis.keys(link_wildcard):
            self.redis.smembers(key)



from sqlalchemy import *
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy.sql.expression import _Null

from demo.util import Repr, UpdatedAt

DaphneBase = declarative_base()

class VlanServiceProvider(DaphneBase, UpdatedAt, Repr):
    __tablename__ = "vlan_service_providers"
    id = Column(Integer, primary_key=True)
    name = Column(String)
    canUpgrade = Column(Boolean)
    children = relationship("VlanServiceConnection", back_populates="parent")
    pending_update = Column(Boolean, default=False, nullable=False)


class VlanServiceConnection(DaphneBase, UpdatedAt, Repr):
    __tablename__ = "vlan_service_connections"
    id = Column(Integer, primary_key=True)
    name = Column(String)
    region = Column(String)
    pop = Column(String)
    externalRef = Column(String)
    parent_id = Column(Integer, ForeignKey("vlan_service_providers.id"))
    parent = relationship("VlanServiceProvider", back_populates="children")



from datetime import datetime

from sqlalchemy import Column, DateTime
from sqlalchemy.orm import MANYTOONE

from sync.mixins import Timestamped


class Repr:
    def __repr__(self):
        columns = self.__table__.columns.values()
        column_key_value = [
            f"{column.name}={getattr(self, column.name)}"
            for column in columns
            if not column.name.endswith("_id")
        ]
        relationship_key_value = [
            f"{relationship[0]}={getattr(self, relationship[0])}"
            for relationship in self.__mapper__.relationships.items()
            if relationship[1].direction == MANYTOONE
        ]
        return f"{self.__class__.__name__}({', '.join([*column_key_value, *relationship_key_value])})"


class UpdatedAt(Timestamped):
    _updated_at = Column(DateTime, default=datetime.utcnow)

    def updated_at(self):
        if not self._updated_at or not self._updated_at.timestamp():
            return 0
        else:
            return self._updated_at.timestamp()

    def update_timestamp(self, timestamp):
        self._updated_at = datetime.utcfromtimestamp(timestamp)



import os
from typing import Callable, List, Type

from redis.client import Redis

from sync.backends.redislinkprovider import RedisLinkProvider
from demo.fixtures import populate_daphne, daphne_sqa_ds, desire_sqa_ds, populate_desire
from demo.celery.app import app
from celery.signals import worker_init
from sync.datasource import DataSource
from sync.mapping import Mapping
from sync.syncer import Syncer


def create_syncer(left_ds, right_ds, mappings):
    redis_link_provider = RedisLinkProvider(host="redis")
    syncer = Syncer(
        left_ds=left_ds,
        right_ds=right_ds,
        mappings=mappings,
        link_provider=redis_link_provider,
    )
    return syncer


@app.task
def sync_all(
    left_ds_producer: Callable[[], DataSource],
    right_ds_producer: Callable[[], DataSource],
    mappings: List[Mapping],
):
    syncer = create_syncer(left_ds_producer(), right_ds_producer(), mappings)
    return syncer.sync_all()


@app.task
def sync_one(
    left_ds_producer: Callable[[], DataSource],
    right_ds_producer: Callable[[], DataSource],
    mappings: List[Mapping],
    type: Type,
    id: str,
):
    syncer = create_syncer(left_ds_producer(), right_ds_producer(), mappings)
    return syncer.sync_entity(type, id)


# initialize desire and daphne
@worker_init.connect
def startup(**kwargs):
    cleanup()
    initialize()


def cleanup():
    Redis(host="redis").flushdb()
    if os.path.exists("/tmp/daphne.sqlite"):
        os.remove("/tmp/daphne.sqlite")
    if os.path.exists("/tmp/desire.sqlite"):
        os.remove("/tmp/desire.sqlite")


def initialize():
    populate_daphne(daphne_sqa_ds())
    populate_desire(desire_sqa_ds())



from dataclasses import dataclass, field
from enum import Enum, auto
from typing import Tuple, Set, Union, Dict


class Mode(Enum):
    LEFT_TO_RIGHT = auto()
    RIGHT_TO_LEFT = auto()


class DirtyMode(Enum):
    DO_NOT_OVERWRITE = auto()
    UPDATE = auto()
    IGNORE = auto()


class AttributeMap:
    def __init__(
            self,
            from_attribute: Union[str, Tuple[str, ...]],
            to_attribute: Union[str, Tuple[str, ...]],
            from_attribute_type_hint: Union[str, Tuple[str, ...]] = None,
            to_attribute_type_hint: Union[str, Tuple[str, ...]] = None,
    ):
        self.from_attribute = from_attribute
        self.to_attribute = to_attribute
        self.from_attribute_type_hint = from_attribute_type_hint
        self.to_attribute_type_hint = to_attribute_type_hint

    def swapped(self):
        return AttributeMap(
            self.to_attribute,
            self.from_attribute,
            self.to_attribute_type_hint,
            self.from_attribute_type_hint,
        )

    def __repr__(self):
        return f"AttributeMap({self.from_attribute}:{self.from_attribute_type_hint},{self.to_attribute}:{self.to_attribute_type_hint})"


@dataclass
class Mapping:
    modes: Set[Mode] = field(default_factory=set)
    entity_types: Tuple[type, type] = field(default_factory=set)
    keys: Set[AttributeMap] = field(default_factory=set)
    attributes: Set[AttributeMap] = field(default_factory=set)
    create: bool = True
    delete_orphaned_entities: bool = False
    dirty_mode: DirtyMode = field(default=DirtyMode.IGNORE)
    dirty_attribute_name: str = "pending_update"
    left_conditions: Dict[str, type] = field(default_factory=dict)
    right_conditions: Dict[str, type] = field(default_factory=dict)

    def swapped(self):
        if Mode.RIGHT_TO_LEFT not in self.modes:
            raise SwapUnsupported()

        return Mapping(
            self.modes,
            (self.entity_types[1], self.entity_types[0]),
            {key.swapped() for key in self.keys},
            {attribute.swapped() for attribute in self.attributes},
            self.create,
            self.delete_orphaned_entities,
            self.dirty_mode,
            self.dirty_attribute_name,
            left_conditions=self.right_conditions,
            right_conditions=self.left_conditions,
        )

    def from_paths(self):
        return {key.from_attribute for key in self.keys}.union(
            {attribute.from_attribute for attribute in self.attributes}
        )


class SwapUnsupported(Exception):
    pass
